@@ -83,6 +83,7 @@ def _wrapOpen(func):
 gzip_open = gzip.open
 if sys.version_info >= (3, 3):
   import lzma
+  lzma_open = lzma.open
   bz2_open = bz2.open
   _read_mode = 'rt'
 else:
@@ -91,6 +92,7 @@ else:
   _read_mode = 'r'
   try:
     from backports import lzma
+    lzma_open = _wrapOpen(lzma.open)
   except ImportError:
     lzma = None
 
@@ -99,18 +101,22 @@ FILE_OPENER_LIST = [
   (bz2_open, IOError),
 ]
 if lzma is not None:
-  FILE_OPENER_LIST.append((lzma.open, lzma.LZMAError))
+  FILE_OPENER_LIST.append((lzma_open, lzma.LZMAError))
 
 # XXX: what encoding ? apache doesn't document one, but requests are supposed
 # to be urlencoded, so pure ascii. Are timestamps localised ?
+# Unlike apache, Caddy does not escape referrer headers, so caddy log files may contain
+# non ascii characters.
+# We read them as ascii, replacing non-ascii characters by unicode replacement character.
 INPUT_ENCODING = 'ascii'
+INPUT_ENCODING_ERROR_HANDLER = 'replace'
 
 MONTH_VALUE_DICT = dict((y, x) for (x, y) in enumerate(('Jan', 'Feb', 'Mar',
   'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'), 1))
 
 US_PER_S = 10 ** 6
 
-N_SLOWEST = 20
+N_HOTTEST_PAGES_DEFAULT = 20
 N_ERROR_URL = 10
 N_REFERRER_PER_ERROR_URL = 5
 N_USER_AGENT = 20
@@ -355,7 +361,9 @@ _APDEXDateDictAsJSONState = lambda date_dict: dict(((y, z.asJSONState())
 
 class GenericSiteStats(object):
   def __init__(self, threshold, getDuration, suffix, error_detail=False,
-      user_agent_detail=False):
+      user_agent_detail=False,
+      # Non-generic parameters
+      **kw):
     self.threshold = threshold
     self.suffix = suffix
     self.error_detail = error_detail
@@ -403,6 +411,7 @@ class GenericSiteStats(object):
       x_min=None, x_max=None,
       apdex_y_min=None, hit_y_min=None, hit_y_max=None,
       apdex_y_scale=None, hit_y_scale=None,
+      n_hottest_pages=N_HOTTEST_PAGES_DEFAULT,
     ):
     result = []
     append = result.append
@@ -417,7 +426,7 @@ class GenericSiteStats(object):
     append(APDEXStats.asHTMLHeader())
     append('<th>url</th></tr>')
     for url, data in sorted(self.url_apdex.iteritems(), key=lambda x: x[1].getAverage() * x[1].hit,
-        reverse=True)[:N_SLOWEST]:
+        reverse=True)[:n_hottest_pages]:
       append('<tr>')
       append(data.asHTML(self.threshold))
       append('<td class="text">%s</td></tr>' % escape(url))
@@ -454,7 +463,7 @@ class GenericSiteStats(object):
     for status, data_dict in sorted(filtered_status.iteritems(),
         key=ITEMGETTER0):
       has_errors |= statusIsError(status)
-      append('<tr><th>%s</th>' % statusAsHtml(status))
+      append('<tr title="%s"><th>%s</th>' % (status, statusAsHtml(status)))
       append(hitTd(sum(data_dict.itervalues()), status))
       for column in column_list:
         append(hitTd(data_dict[column], status))
@@ -554,16 +563,26 @@ class ERP5SiteStats(GenericSiteStats):
     count line as belonging to a document of that module
   """
   def __init__(self, threshold, getDuration, suffix, error_detail=False,
-      user_agent_detail=False):
+      user_agent_detail=False, erp5_expand_other=False):
     super(ERP5SiteStats, self).__init__(threshold, getDuration, suffix,
       error_detail=error_detail, user_agent_detail=user_agent_detail)
+
+    self.expand_other = erp5_expand_other
+
     # Key levels:
     # - module id (string)
     # - is document (bool)
     # - date (string)
     self.module = defaultdict(partial(defaultdict, partial(
       defaultdict, partial(APDEXStats, threshold, getDuration))))
-    self.no_module = defaultdict(partial(APDEXStats, threshold, getDuration))
+
+    # Key levels:
+    # - id (string)
+    #   => 'other' only if expand_other == False
+    # - date (string)
+    self.no_module = defaultdict(partial(
+        defaultdict, partial(APDEXStats, threshold, getDuration)))
+
     self.site_search = defaultdict(partial(APDEXStats, threshold, getDuration))
 
   def rescale(self, convert, getDuration):
@@ -575,11 +594,17 @@ class ERP5SiteStats(GenericSiteStats):
         for value_date, data in date_dict.iteritems():
           new_date_dict[convert(value_date)].accumulateFrom(data)
         document_dict[is_document] = new_date_dict
-    for attribute_id in ('no_module', 'site_search'):
-      attribute = defaultdict(partial(APDEXStats, threshold, getDuration))
-      for value_date, data in getattr(self, attribute_id).iteritems():
-        attribute[convert(value_date)].accumulateFrom(data)
-      setattr(self, attribute_id, attribute)
+
+    for id_, date_dict in self.no_module.iteritems():
+      new_date_dict = defaultdict(partial(APDEXStats, threshold, getDuration))
+      for value_date, data in date_dict.iteritems():
+        new_date_dict[convert(value_date)].accumulateFrom(data)
+      self.no_module[id_] = new_date_dict
+
+    attribute = defaultdict(partial(APDEXStats, threshold, getDuration))
+    for value_date, data in getattr(self, attribute_id).iteritems():
+      attribute[convert(value_date)].accumulateFrom(data)
+    self.site_search = attribute
 
   def accumulate(self, match, url_match, value_date):
     split = self.suffix(url_match.group('url')).split('?', 1)[0].split('/')
@@ -592,13 +617,16 @@ class ERP5SiteStats(GenericSiteStats):
     elif split and split[0] == 'ERP5Site_viewSearchResult':
       super(ERP5SiteStats, self).accumulate(match, url_match, value_date)
       self.site_search[value_date].accumulate(match)
+    elif split and self.expand_other:
+      self.no_module[split[0]][value_date].accumulate(match)
     else:
-      self.no_module[value_date].accumulate(match)
+      self.no_module['other'][value_date].accumulate(match)
 
   def asHTML(self, date_format, placeholder_delta, graph_period, graph_coefficient,
       encoding, stat_filter=lambda x: x, x_min=None, x_max=None,
       apdex_y_min=None, hit_y_min=None, hit_y_max=None,
       apdex_y_scale=None, hit_y_scale=None,
+      n_hottest_pages=N_HOTTEST_PAGES_DEFAULT,
     ):
     result = []
     append = result.append
@@ -609,10 +637,16 @@ class ERP5SiteStats(GenericSiteStats):
       None))
     filtered_module = defaultdict(partial(defaultdict, partial(
       defaultdict, partial(APDEXStats, self.threshold, None))))
-    filtered_no_module = defaultdict(partial(APDEXStats, self.threshold, None))
-    for value_date, value in self.no_module.iteritems():
-      filtered_no_module[stat_filter(value_date)].accumulateFrom(value)
-    column_set = set(filtered_no_module)
+    other_overall = APDEXStats(self.threshold, None)
+    filtered_no_module = defaultdict(partial(
+      defaultdict, partial(APDEXStats, self.threshold, None)))
+    column_set = set()
+    for key, data_dict in self.no_module.iteritems():
+      filtered_id_dict = filtered_no_module[key]
+      for value_date, value in data_dict.iteritems():
+        filtered_id_dict[stat_filter(value_date)].accumulateFrom(value)
+        other_overall.accumulateFrom(value)
+      column_set.update(filtered_id_dict)
     filtered_site_search = defaultdict(partial(APDEXStats, self.threshold,
       None))
     for value_date, value in self.site_search.iteritems():
@@ -673,26 +707,30 @@ class ERP5SiteStats(GenericSiteStats):
       append('</td>')
     for module_id, data_dict in sorted(filtered_module.iteritems(),
         key=ITEMGETTER0):
-      append('<tr class="group_top"><th rowspan="2">%s</th>'
-        '<th>module</th>' % module_id)
+      append('<tr class="group_top" title="%s (module)"><th rowspan="2">%s</th>'
+        '<th>module</th>' % (module_id, module_id))
       hiddenGraph(self.module[module_id][False], module_id + ' (module)')
       apdexAsColumns(data_dict[False])
-      append('</tr><tr class="group_bottom"><th>document</th>')
+      append('</tr><tr class="group_bottom" title="%s (document)"><th>document</th>' % module_id)
       hiddenGraph(self.module[module_id][True], module_id + '  (document)')
       apdexAsColumns(data_dict[True])
       append('</tr>')
-    append('<tr class="group_top group_bottom"><th colspan="2">site search'
+    append('<tr class="group_top group_bottom" title="site search"><th colspan="2">site search'
       '</th>')
     hiddenGraph(self.site_search, 'site search')
     site_search_overall = apdexAsColumns(filtered_site_search)
-    append('<tr class="group_top group_bottom"><th colspan="2">other</th>')
-    hiddenGraph(self.no_module, 'other')
-    no_module_overall = apdexAsColumns(filtered_no_module)
-    append('</tr></table><h2>Per-level overall</h2><table class="stats"><tr>'
+    append('</tr>')
+    for id_, date_dict in sorted(filtered_no_module.iteritems()):
+      append('<tr class="group_top group_bottom" title="%s"><th colspan="2">%s</th>'
+             % (id_, id_))
+      hiddenGraph(self.no_module[id_], id_)
+      apdexAsColumns(date_dict)
+      append('</tr>')
+    append('</table><h2>Per-level overall</h2><table class="stats"><tr>'
       '<th>level</th>')
     append(APDEXStats.asHTMLHeader())
     append('</tr><tr><th>other</th>')
-    append(no_module_overall.asHTML(self.threshold))
+    append(other_overall.asHTML(self.threshold))
     append('</tr><tr><th>site search</th>')
     append(site_search_overall.asHTML(self.threshold))
     append('</tr><tr><th>module</th>')
@@ -707,6 +745,7 @@ class ERP5SiteStats(GenericSiteStats):
       apdex_y_min=apdex_y_min, hit_y_min=hit_y_min, hit_y_max=hit_y_max,
       apdex_y_scale=apdex_y_scale,
       hit_y_scale=hit_y_scale,
+      n_hottest_pages=n_hottest_pages,
     ))
     return '\n'.join(result)
 
@@ -719,10 +758,17 @@ class ERP5SiteStats(GenericSiteStats):
         date_dict = module_dict[is_document == 'true']
         for value_date, apdex_state in date_dict_state.iteritems():
           date_dict[value_date] = APDEXStats.fromJSONState(apdex_state, getDuration)
-    for attribute_id in ('no_module', 'site_search'):
-      attribute = getattr(result, attribute_id)
-      for value_date, apdex_state in state[attribute_id].iteritems():
-        attribute[value_date] = APDEXStats.fromJSONState(apdex_state, getDuration)
+
+    for id_, date_dict in state['no_module'].iteritems():
+      no_module_dict = result.no_module[id_]
+      for value_date, apdex_state in date_dict.iteritems():
+        no_module_dict[value_date] = APDEXStats.fromJSONState(
+            apdex_state, getDuration)
+
+    for value_date, apdex_state in state['site_search'].iteritems():
+      result.site_search[value_date] = APDEXStats.fromJSONState(
+        apdex_state, getDuration)
+
     return result
 
   def asJSONState(self):
@@ -732,9 +778,12 @@ class ERP5SiteStats(GenericSiteStats):
       module_dict_state = module[module_id] = {}
       for is_document, date_dict in module_dict.iteritems():
         module_dict_state[is_document] = _APDEXDateDictAsJSONState(date_dict)
-    for attribute_id in ('no_module', 'site_search'):
-      result[attribute_id] = _APDEXDateDictAsJSONState(getattr(self,
-        attribute_id))
+
+    result['no_module'] = no_module = {}
+    for id_, date_dict in self.no_module.iteritems():
+      no_module[id_] = _APDEXDateDictAsJSONState(date_dict)
+
+    result['site_search'] = _APDEXDateDictAsJSONState(self.site_search)
     return result
 
   def accumulateFrom(self, other):
@@ -746,10 +795,15 @@ class ERP5SiteStats(GenericSiteStats):
         date_dict = module_dict[is_document]
         for value_date, apdex in other_date_dict.iteritems():
           date_dict[value_date].accumulateFrom(apdex)
-    for attribute_id in ('no_module', 'site_search'):
-      attribute = getattr(self, attribute_id)
-      for value_date, apdex in getattr(other, attribute_id).iteritems():
-        attribute[value_date].accumulateFrom(apdex)
+
+    for id_, other_date_dict in other.no_module.iteritems():
+      date_dict = self.no_module[id_]
+      for value_date, apdex in other_date_dict.iteritems():
+        date_dict.accumulateFrom(apdex)
+
+    attribute = self.site_search
+    for value_date, apdex in other.site_search.iteritems():
+      attribute[value_date].accumulateFrom(apdex)
 
 DURATION_US_FORMAT = '%D'
 DURATION_S_FORMAT = '%T'
@@ -1100,6 +1154,7 @@ def asHTML(out, encoding, per_site, args, default_site, period_parameter_dict,
   for hit_date, hit in sorted(hit_per_day.iteritems(), key=ITEMGETTER0):
     out.write('<tr><td>%s</td><td>%s</td></tr>' % (hit_date, hit))
   out.write('</table>')
+  n_hottest_pages = args.n_hottest_pages
   for i, (site_id, data) in site_list:
     out.write('<h1 id="%s" title="%s">%s</h1>' % (i, escape(repr(site_id),
       quote=True), html_site_caption_dict[site_id]))
@@ -1130,6 +1185,7 @@ def asHTML(out, encoding, per_site, args, default_site, period_parameter_dict,
       apdex_y_min=apdex_y_min, hit_y_min=hit_y_min, hit_y_max=hit_y_max,
       apdex_y_scale=apdex_y_scale,
       hit_y_scale=hit_y_scale,
+      n_hottest_pages=n_hottest_pages,
     ))
   end_stat_time = time.time()
   if args.stats:
@@ -1254,6 +1310,8 @@ def main():
     help='Include detailed report (url & referers) for error statuses.')
   group.add_argument('-u', '--user-agent-detail', action='store_true',
     help='Include report of most frequent user agents.')
+  group.add_argument('--erp5-expand-other', action='store_true',
+    help='Expand ERP5 `other` stats')
 
   group.add_argument('-f', '--format', choices=format_generator,
     default='html', help='Format in which output should be generated.')
@@ -1277,6 +1335,9 @@ def main():
   group.add_argument('--hit-yscale', default='linear',
     choices=hit_y_scale_dict,
     help='hit graph ordinate scale. Default: %(default)s')
+  group.add_argument('--n-hottest-pages', type=int,
+    default=N_HOTTEST_PAGES_DEFAULT,
+    help='Number of hottest pages to display.')
 
   group = parser.add_argument_group('site matching', 'Earlier arguments take '
     'precedence. Arguments are Python regexes, matching urlencoded strings.'
@@ -1398,6 +1459,7 @@ def main():
   threshold = args.apdex
   error_detail = args.error_detail
   user_agent_detail = args.user_agent_detail
+  erp5_expand_other = args.erp5_expand_other
   file_count = len(infile_list)
   per_site = {}
   if '-' in args.state_file and '-' in infile_list:
@@ -1452,7 +1514,7 @@ def main():
       logfile = sys.stdin
     else:
       for opener, exc in FILE_OPENER_LIST:
-        logfile = opener(filename, _read_mode, encoding=INPUT_ENCODING)
+        logfile = opener(filename, _read_mode, encoding=INPUT_ENCODING, errors=INPUT_ENCODING_ERROR_HANDLER)
         try:
           logfile.readline()
         except exc:
@@ -1461,7 +1523,7 @@ def main():
           logfile.seek(0)
           break
       else:
-        logfile = codecs.open(filename, _read_mode, encoding=INPUT_ENCODING)
+        logfile = codecs.open(filename, _read_mode, encoding=INPUT_ENCODING, errors=INPUT_ENCODING_ERROR_HANDLER)
     lineno = 0
     for lineno, line in enumerate(logfile, 1):
       if show_progress and lineno % 5000 == 0:
@@ -1530,7 +1592,8 @@ def main():
         site_data = per_site[site]
       except KeyError:
         site_data = per_site[site] = action(threshold, getDuration,
-          error_detail=error_detail, user_agent_detail=user_agent_detail)
+          error_detail=error_detail, user_agent_detail=user_agent_detail,
+          erp5_expand_other=erp5_expand_other)
       try:
         site_data.accumulate(match, url_match, hit_date)
       except Exception:
